#!/usr/bin/env scheme-script
;; -*- mode: scheme; coding: utf-8 -*- !#
;; Copyright © 2018 Göran Weinholt <goran@weinholt.se>
;; SPDX-License-Identifier: GPL-3.0-or-later

#!r6rs

(import
  (rnrs (6))
  (industria hexdump)
  (struct pack)
  (industria crypto rsa)
  (terrible-tls simple)
  (prefix (terrible-tls x509) x509:)
  (MumbleProto ACL)
  (MumbleProto Authenticate)
  (MumbleProto Authenticate)
  (MumbleProto BanList)
  (MumbleProto ChannelRemove)
  (MumbleProto ChannelState)
  (MumbleProto ChannelState)
  (MumbleProto CodecVersion)
  (MumbleProto ContextAction)
  (MumbleProto ContextActionModify)
  (MumbleProto CryptSetup)
  (MumbleProto PermissionDenied)
  (MumbleProto PermissionQuery)
  (MumbleProto Ping)
  (MumbleProto QueryUsers)
  (MumbleProto Reject)
  (MumbleProto Reject)
  (MumbleProto RequestBlob)
  (MumbleProto ServerConfig)
  (MumbleProto ServerConfig)
  (MumbleProto ServerSync)
  (MumbleProto ServerSync)
  (MumbleProto SuggestConfig)
  (MumbleProto TextMessage)
  (MumbleProto UDPTunnel)
  (MumbleProto UserList)
  (MumbleProto UserRemove)
  (MumbleProto UserState)
  (MumbleProto UserStats)
  (MumbleProto Version)
  (MumbleProto Version)
  (MumbleProto VoiceTarget))

;; Packet types
(define Version 0)
(define UDPTunnel 1)
(define Authenticate 2)
(define Ping 3)
(define Reject 4)
(define ServerSync 5)
(define ChannelRemove 6)
(define ChannelState 7)
(define UserRemove 8)
(define UserState 9)
(define BanList 10)
(define TextMessage 11)
(define PermissionDenied 12)
(define ACL 13)
(define QueryUsers 14)
(define CryptSetup 15)
(define ContextActionModify 16)
(define ContextAction 17)
(define UserList 18)
(define VoiceTarget 19)
(define PermissionQuery 20)
(define CodecVersion 21)
(define UserStats 22)
(define RequestBlob 23)
(define ServerConfig 24)
(define SuggestConfig 25)

(define (start-connection outp username password)
  (let ((verc (make-Version-builder)))
    (set-Version-builder-version! verc #x00010200) ;1.2.0
    (set-Version-builder-release! verc "0.1.0")
    (set-Version-builder-os! verc "Scheme")
    (set-Version-builder-os_version! verc "R6RS")
    (let ((payload (call-with-bytevector-output-port
                     (lambda (p)
                       (Version-write (Version-builder-build verc) p)))))
      (put-pack outp "!uSL" Version (bytevector-length payload))
      (put-bytevector outp payload)))

  (let ((c (make-Authenticate-builder)))
    (set-Authenticate-builder-username! c username)
    (unless (string=? password "")
      (set-Authenticate-builder-password! c password))
    (set-Authenticate-builder-tokens! c '())
    (let ((payload (call-with-bytevector-output-port
                     (lambda (p)
                       (Authenticate-write (Authenticate-builder-build c) p)))))
      (put-pack outp "!uSL" Authenticate (bytevector-length payload))
      (put-bytevector outp payload)))
  (flush-output-port outp))

(define (get-packet inp)
  (when (port-eof? inp)
    (error 'get-packet "End of file" inp))
  (let-values (((type len) (get-unpack inp "!uSL")))
    (let ((payload (get-bytevector-n inp len)))
      (when (bytevector? payload)
        (newline)
        (display type)
        (newline)
        (hexdump (current-output-port) payload))
      (unless (and (bytevector? payload)
                   (eqv? (bytevector-length payload) len))
        (error 'get-packet "Short read"))
      (values type payload))))

(define (main username password target-channel hostname port cert-fn key-fn)
  (let-values (((inp outp session)
                (tls-connect hostname
                             port
                             (if (string=? cert-fn "")
                                 '()
                                 (x509:certificates-from-pem-file cert-fn))
                             (if (string=? key-fn "")
                                 #f
                                 (rsa-private-key-from-pem-file key-fn)))))
    (define channel-by-name (make-hashtable string-hash string=?))
    (let-values (((type payload) (get-packet inp)))
      (cond
        ((= type Version)
         (let ((pkt (Version-read (open-bytevector-input-port payload))))
           (write (list (number->string (Version-version pkt) 16)
                        (Version-release pkt)
                        (Version-os pkt)
                        (Version-os_version pkt)))
           (newline))
         (start-connection outp username password))
        (else
         (error 'mummel-connect "Unexpected packet during startup" type))))

    (let lp ()
      (flush-output-port outp)
      (let-values (((type payload) (get-packet inp)))
        (cond
          ((= type Reject)
           (let ((pkt (Reject-read (open-bytevector-input-port payload))))
             (write (list (and (has-Reject-type? pkt) (Reject-type pkt))
                          (and (has-Reject-reason? pkt) (Reject-reason pkt))))
             (newline)))

          ((= type ChannelState)
           (let ((pkt (ChannelState-read (open-bytevector-input-port payload))))
             (let ((id (and (has-ChannelState-channel_id? pkt) (ChannelState-channel_id pkt)))
                   (name (and (has-ChannelState-name? pkt) (ChannelState-name pkt))))
               (write (list id name))
               (newline)
               (when (and name id)
                 (hashtable-set! channel-by-name name id))
               (lp))))

          ((= type ServerConfig)
           (display "ServerConfig")
           (newline)
           (let ((pkt (ServerConfig-read (open-bytevector-input-port payload))))
             (let ((welcome-text (and (has-ServerConfig-welcome_text? pkt)
                                      (ServerConfig-welcome_text pkt))))
               (write welcome-text)
               (newline)))
           (lp))

          ((= type ServerSync)
           (let ((pkt (ServerSync-read (open-bytevector-input-port payload))))
             (let ((welcome-text (and (has-ServerSync-welcome_text? pkt)
                                      (ServerSync-welcome_text pkt))))
               (write welcome-text)
               (newline)))

           (let ((id (hashtable-ref channel-by-name target-channel #f)))
             (when id
               (let ((req (make-ChannelState-builder)))
                 (set-ChannelState-builder-channel_id! req id)
                 (set-ChannelState-builder-description! req
                                                        (get-string-all (current-input-port)))
                 (let ((payload (call-with-bytevector-output-port
                                  (lambda (p)
                                    (ChannelState-write (ChannelState-builder-build req) p)))))
                   (put-pack outp "!uSL" ChannelState (bytevector-length payload))
                   (put-bytevector outp payload)
                   (lp))))))

          ((= type PermissionDenied)
           (display "Permission denied\n")
           (lp))
          (else
           (lp)))))))

(apply
 (case-lambda
   ((username password target-channel hostname port cert-fn key-fn)
    (main username password target-channel hostname port cert-fn key-fn))
   (else
    (display "mummel.sps username password target-channel hostname 64738 cert.pem cert.key\n"
             (current-error-port))
    (exit 1)))
 (cdr (command-line)))
